#!/usr/bin/env python3
"""
Lightweight QC analysis and reporting for daily phantom scans

Authors
----
Mike Tyszka, Caltech Brain Imaging Center

MIT License

Copyright (c) 2019 Mike Tyszka

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys
import argparse
import pkg_resources
import subprocess

from cbicqc.cbicqc import CBICQC


def main():
    # Parse command line arguments
    parser = argparse.ArgumentParser(description='Lightweight daily phantom QC analysis and reporting')
    parser.add_argument('-d', '--dir', default='.', help='BIDS QC dataset directory')
    parser.add_argument('-m', '--mode', default='phantom', help="QC Mode (phantom or live)")
    parser.add_argument('-p', '--past', default=12, type=int, help='Number of past months to summarize [12]')
    parser.add_argument('--sub', default='', help='Subject ID')
    parser.add_argument('--ses', default='', help='Session ID')

    # Parse command line arguments
    args = parser.parse_args()
    bids_dir = os.path.realpath(args.dir)
    subj_id = args.sub
    sess_id = args.ses
    mode = args.mode
    past_months = args.past

    # Read version from setup.py
    ver = pkg_resources.get_distribution('cbicqc').version

    # Splash
    print('')
    print('-----------------------------')
    print('CBIC Quality Control Analysis')
    print('-----------------------------')
    print('Version : {}'.format(ver))
    print('')
    print('BIDS Directory : {}'.format(bids_dir))
    print('Subject : {}'.format(subj_id if len(subj_id) > 0 else 'All Subjects'))
    print('Session : {}'.format(sess_id if len(sess_id) > 0 else 'All Sessions'))

    # Long term summary relevent for phantom QC
    if 'phantom' in mode:
        print('Summary : {} months'.format(past_months))

    # Setup QC analysis
    qc = CBICQC(bids_dir=bids_dir, subject=subj_id, session=sess_id, mode=mode, past_months=past_months)

    # Run analysis
    qc.run()

    # Clean exit
    sys.exit(0)


# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Create summary PDF report

AUTHORS
----
Mike Tyszka, Ph.D., Caltech Brain Imaging Center

MIT License

Copyright (c) 2020 Mike Tyszka

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import shutil
import tempfile
import numpy as np
from datetime import datetime

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

from sklearn.cluster import DBSCAN
from sklearn.preprocessing import StandardScaler

from pandas.plotting import register_matplotlib_converters
from reportlab.lib.enums import TA_JUSTIFY
from reportlab.lib.pagesizes import letter
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.units import inch
from reportlab.platypus import (SimpleDocTemplate,
                                Paragraph,
                                Spacer,
                                Image,
                                Table,
                                PageBreak)

from .graphics import (metric_trend_plot)


class Summarize:

    def __init__(self, report_dir, metrics_df, past_months):
        """
        Create summary report PDF and CSV file for all sessions

        :param report_dir: str, report output directory in derivatives
        :param metrics_df: DataFrame, session metric dataframe
        :param past_months: int, number of past months to summarize
        """

        # For datetime axis labeling without warnings
        register_matplotlib_converters()

        # Add plot-friendly Date column
        metrics_df['Date'] = [datetime.strptime(dt, '%Y-%m-%dT%H:%M:%S.%f')
                              for dt in metrics_df['AcquisitionDateTime'].values]

        # Metrics of interest to plot and save to CSV
        self._metric_names = [
            'SNR',
            'SFNR',
            'NoiseFloor',
            'Drift',
            'NyquistSpikes',
            'AirSpikes'
        ]

        self._report_dir = report_dir
        self._metrics_df = metrics_df
        self._subject = metrics_df['Subject'][0]
        self._past_months = past_months
        self._summary_pdf = os.path.join(report_dir, '{}_summary.pdf'.format(self._subject))
        self._summary_csv = self._summary_pdf.replace('.pdf', '.csv')

        # Identify outliers before generating plots and writing CSV
        self._outliers()

        #
        # Summary PDF construction
        #

        # Create working directory for images
        self._work_dir = tempfile.mkdtemp()

        # Contents - list of flowables to be built into a document
        self._contents = []

        # Add a justified paragraph style
        self._pstyles = getSampleStyleSheet()
        self._pstyles.add(ParagraphStyle(name='Justify', alignment=TA_JUSTIFY))

        self._init_pdf()
        self._add_coverpage()
        self._add_metric_graphs()

        self._doc.build(self._contents)

        # Delete working directory
        shutil.rmtree(self._work_dir)

        # Finally write CSV for metrics of interest
        self._write_csv()

    def _init_pdf(self):

        # Create a new PDF document
        self._doc = SimpleDocTemplate(self._summary_pdf,
                                      pagesize=letter,
                                      rightMargin=0.5 * inch,
                                      leftMargin=0.5 * inch,
                                      topMargin=0.5 * inch,
                                      bottomMargin=0.5 * inch)

    def _add_coverpage(self):

        ptext = '<font size=24>CBIC Quality Control Summary</font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.5 * inch))

        timestamp = datetime.now().strftime('%Y-%m-%d at %H:%M:%S')
        ptext = '<font size=12>Generated by CBICQC on {}</font>'.format(timestamp)
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        ptext = '<font size=14><b>Subject Details</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        # First session metadata - assume identical for all sessions
        m = self._metrics_df.iloc[0]

        meta = [['Subject', self._subject],
                ['Scanner', m['StationName'] + ' ' + m['DeviceSerialNumber']],
                ['Software Version', m['SoftwareVersions']],
                ['Coil Name', m['ReceiveCoilName']]]

        self._contents.append(Table(meta, hAlign='LEFT'))
        self._contents.append(Spacer(1, 0.25 * inch))

        ptext = '<font size=14><b>Potential Outlier Sessions</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        ptext = '<font size=12>Identified by conservative DBSCAN clustering (epsilon = 1.5)</font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        df = self._metrics_df[['Date', 'Outlier']]
        meta = df[df['Outlier'] == 'Outlier'].values.tolist()
        self._contents.append(Table(meta, hAlign='LEFT'))
        self._contents.append(Spacer(1, 0.25 * inch))

    def _add_metric_graphs(self):
        """
        Add timecourse and histogram plots for the given metric class ('Noise' or 'Spike/Drift')
        :return:
        """

        # Page break
        self._contents.append(PageBreak())

        ptext = '<font size=14><b>Session Metric Trends</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        trend_png_fname = self._plot_trends()
        trends_img = Image(trend_png_fname, 7.0 * inch, 9.0 * inch, hAlign='LEFT')
        self._contents.append(trends_img)

    def _plot_trends(self):
        """
        Generate a PNG of the trends and histogram for each of the passed metrics
        :return:
        """

        # Number of metrics to plot
        n_metrics = len(self._metric_names)

        # Output PNG filename
        png_fname = os.path.join(self._work_dir, 'metric_trends.png')

        # Setup plot grid
        plt.figure(figsize=(14, 18))
        gs = gridspec.GridSpec(n_metrics, 2, width_ratios=[3, 1])

        # Fill each of the subplots
        for mc, m_name in enumerate(self._metric_names):
            metric_trend_plot(mc, m_name, self._metrics_df, gridspec=gs, past_months=self._past_months)

        # Tweak subplot margins and spacing
        plt.tight_layout()

        # Save plot to file
        plt.savefig(png_fname, dpi=300)

        return png_fname

    def _write_csv(self):

        # Full column list including subject, session, date and metrics
        cols_to_write = ['Subject', 'Session', 'Date'] + self._metric_names

        print()
        print('Writing metrics of interest to {:s}'.format(self._summary_csv))

        self._metrics_df.to_csv(self._summary_csv,
                                columns=cols_to_write,
                                header=True,
                                index=False)

    def _outliers(self):
        """
        Identify within sample outliers using DBSCAN
        Adds outlier flag column to metric DataFrame

        :return:
        """

        n = len(self._metrics_df)

        # DBSCAN clustering
        X = np.array(self._metrics_df[self._metric_names])

        # Standardize mean, sd to 0, 1
        X = StandardScaler().fit_transform(X)

        # Cluster using DBSCAN
        clustering = DBSCAN(eps=1.5, min_samples=5).fit(X)

        # Add boolean Outlier column to DataFrame
        self._metrics_df['Outlier'] = ['Outlier' if l < 0 else 'Inlier' for l in clustering.labels_]


# !/usr/bin/env python
"""
CBICQC quality metrics

AUTHOR : Mike Tyszka
PLACE  : Caltech
DATES  : 2019-06-04 JMT From scratch

This file is part of CBICQC.

   CBICQC is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   CBICQC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
  along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2019 California Institute of Technology.
"""

import os
import subprocess
import numpy as np
import nibabel as nb
from scipy.ndimage.measurements import center_of_mass
from scipy.ndimage import shift
from scipy.spatial.transform import Rotation


def moco_phantom(img_nii):
    """
    Spherical QC phantom requires simpler registration approach.
    Use center of mass registration only.

    :param img_nii: Nifti object,
        4D QC time series
    :return moco_nii: Nifti object,
        Motion corrected 4D QC time series
    :return moco_pars: array,
        Motion parameter array (nt x 6)
    """

    img = img_nii.get_data()
    nt = img.shape[3]
    vox_mm = img_nii.header.get('pixdim')[1:4]

    # Clip intensity range to 1st, 99th percentile for robust CoM
    p1, p99 = np.percentile(img, (1, 99))
    img_clip = np.clip(img, p1, p99)

    moco_img = img.copy()
    moco_pars = np.zeros([nt, 6])

    # Reference center of mass
    com_0 = np.array(center_of_mass(img_clip[:, :, :, 0]))

    for tc in range(1, nt):

        # Center of mass of current volume
        com_t = np.array(center_of_mass(img_clip[:, :, :, tc]))

        # Center of mass shift required to register current and zeroth volumes
        com_d = com_0 - com_t

        # Translate with spline interpolation
        # Use 'nearest neighbor' mode to minimize motion x signal artifacts at image edges
        moco_img[:, :, :, tc] = shift(img[:, :, :, tc], com_d, mode='nearest')

        # Save CoM translation
        # FSL MCFLIRT convention: [rx, ry, rz, dx, dy, dz]
        moco_pars[tc, 3:6] = com_d * vox_mm

    # Create motion corrected Nifti volume
    moco_nii = nb.Nifti1Image(moco_img, img_nii.affine)

    return moco_nii, moco_pars


def moco_live(img_nii, work_dir):
    """
    MCFLIRT-based rigid body motion correction

    :param img_nii: Nifti, image object
    :param work_dir: str, working directory
    :return moco_nii: Nifti, motion corrected image object
    :return moco_pars: array, motion parameter timeseries
    """

    in_fname = os.path.join(work_dir, 'qc.nii.gz')
    out_stub = os.path.join(work_dir, 'qc_mcf')

    # Save QC timeseries for MCFLIRT
    nb.save(img_nii, in_fname)

    mcflirt_cmd = os.path.join(os.environ['FSLDIR'], 'bin', 'mcflirt')

    # Check that MCFLIRT binary exists
    if os.path.isfile(mcflirt_cmd):
        subprocess.run([mcflirt_cmd,
                        '-in', in_fname,
                        '-out', out_stub,
                        '-plots'])
    else:
        print('* MCFLIRT not available - please install FSL and update your environment')

    # Load motion corrected QC timeseries
    moco_nii = nb.load(out_stub + '.nii.gz')
    moco_pars = np.genfromtxt(out_stub + '.par')

    return moco_nii, moco_pars


def total_rotation(rot):
    """
    FSL defaults to an Euler angle rotation representation: R = Rx * Ry * Rz
    The parameter order in the MCFLIRT .par output is [Rx Ry Rz Dx Dy Dz]
    See FSL mcflirt/mcflirt.cc, mcflirt/Globaloptions.h and mathsutils/mathsutils.cc

    :param rot: array, axis degree rotations for each volume (nt x 3) [Rx, Ry, Rz]
    :return: phi_tot: array, total degree rotation angles (nt x 1)
    """

    # Convert axis rotations from degrees to radians
    rot *= np.pi / 180.0

    nt = rot.shape[0]

    phi_tot = np.zeros([nt,])

    for tc in range(0, nt):

        phix, phiy, phiz = rot[tc, :]

        # Individual axis rotation matrices
        Rx = Rotation.from_rotvec([phix, 0, 0])
        Ry = Rotation.from_rotvec([0, phiy, 0])
        Rz = Rotation.from_rotvec([0, 0, phiz])

        # Compose rotations
        Rtot = Rx * Ry * Rz

        # Total rotation from norm of rotation vector (see scipy definition)
        phi_tot[tc] = np.linalg.norm(Rtot.as_rotvec()) * 180.0 / np.pi

    return phi_tot















#!/usr/bin/env python3
"""
Rename and organize Horos QC exported data in <BIDS Root>/incoming and place in <BIDS Root>/sourcedata

AUTHOR
----
Mike Tyszka, Ph.D.

MIT License

Copyright (c) 2019 Mike Tyszka

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys
from glob import glob
import argparse
from pathlib import Path
import pydicom
from shutil import rmtree


def main():

    parser = argparse.ArgumentParser(description='Fix subject and session directory naming in Horos output')

    parser.add_argument('-d', '--dataset', default='.',
                        help='BIDS dataset directory containing sourcedata subdirectory')

    # Parse command line arguments
    args = parser.parse_args()
    dataset_dir = os.path.realpath(args.dataset)

    incoming_dir = os.path.join(dataset_dir, 'incoming')
    sourcedata_dir = os.path.join(dataset_dir, 'sourcedata')
    qc_dir = os.path.join(sourcedata_dir, 'QC')

    # Create single QC subject
    print("Checking that QC subject exists in sourcedata")
    if os.path.isdir(qc_dir):
        print("  It does - continuning")
    else:
        print("  QC subject does not exist - creating QC subject in sourcedata")
        os.makedirs(qc_dir, exist_ok=True)

    # Loop over all Qc study directories in sourcedata
    # Expect subject/session directory names in the form "Qc_<session ID>_*/<session dir>/"
    # Move session subdirectories from Qc_*/<session dir> to Qc/<ScanDate>

    print("Scanning for incoming QC studies")

    for inc_qc_dir in glob(os.path.join(incoming_dir, 'Qc*')):

        print("")
        print("  Processing {}".format(inc_qc_dir))

        # There should be only one session subdirectory
        dlist = list(glob(os.path.join(inc_qc_dir, '*')))

        if len(dlist) > 0:

            ses_dir = dlist[0]

            # Get first DICOM file in ses_dir at any level
            first_dcm = str(list(Path(ses_dir).rglob("*.dcm"))[0])

            # Get acquisition date from DICOM header
            acq_date = acquisition_date(first_dcm)

            # Destination session directory name in QC subject folder
            dest_dir = os.path.join(qc_dir, acq_date)

            # Move and rename session subdirectory
            print('  Moving %s to %s' % (ses_dir, dest_dir))
            os.rename(ses_dir, dest_dir)

        # Delete incoming Qc_* directory
        print('  Deleting %s' % inc_qc_dir)
        rmtree(inc_qc_dir)


def acquisition_date(dcm_fname):
    """
    Extract acquisition date from DICOM header
    :param dcm_fname: DICOM filename
    :return acq_date: str, acquisition date (YYYYMMDD)
    """

    # Default return date
    acq_date = '19010101'

    if not os.path.isfile(dcm_fname):
        print('* File not found - %s' % dcm_fname)

    try:
        ds = pydicom.read_file(dcm_fname, force=True)
    except IOError:
        print("* Problem opening %s" % dcm_fname)
        raise
    except AttributeError:
        print("* Problem opening %s" % dcm_fname)
        raise

    if ds:
        acq_date = ds.AcquisitionDate
    else:
        print('* DICOM header problem - returning %s' % acq_date)

    return acq_date


if 'main' in __name__:

    main()


#!/usr/bin/env python
#
# WORK IN PROGRESS - DO NOT USE!!
#
# DICOM server query and retrieve class
#
# AUTHOR : Mike Tyszka
# PLACE  : Caltech
# DATES  : 2019-05-17 JMT From scratch
#
# This file is part of CBICQC.
#
#    CBICQC is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    CBICQC is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#   along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2019 California Institute of Technology.

import os
import sys
import json

from pydicom.dataset import Dataset
from pynetdicom import AE

from pynetdicom import AE, StoragePresentationContexts
from pynetdicom.sop_class import _QR_CLASSES as qrc


"""
_QR_CLASSES = {
    'PatientRootQueryRetrieveInformationModelFind' : '1.2.840.10008.5.1.4.1.2.1.1',
    'PatientRootQueryRetrieveInformationModelMove' : '1.2.840.10008.5.1.4.1.2.1.2',
    'PatientRootQueryRetrieveInformationModelGet' : '1.2.840.10008.5.1.4.1.2.1.3',
    'StudyRootQueryRetrieveInformationModelFind' : '1.2.840.10008.5.1.4.1.2.2.1',
    'StudyRootQueryRetrieveInformationModelMove' : '1.2.840.10008.5.1.4.1.2.2.2',
    'StudyRootQueryRetrieveInformationModelGet' : '1.2.840.10008.5.1.4.1.2.2.3',
    'PatientStudyOnlyQueryRetrieveInformationModelFind' : '1.2.840.10008.5.1.4.1.2.3.1',
    'PatientStudyOnlyQueryRetrieveInformationModelMove' : '1.2.840.10008.5.1.4.1.2.3.2',
    'PatientStudyOnlyQueryRetrieveInformationModelGet' : '1.2.840.10008.5.1.4.1.2.3.3',
    'CompositeInstanceRootRetrieveMove' : '1.2.840.10008.5.1.4.1.2.4.2',
    'CompositeInstanceRootRetrieveGet' : '1.2.840.10008.5.1.4.1.2.4.3',
    'CompositeInstanceRetrieveWithoutBulkDataGet' : '1.2.840.10008.5.1.4.1.2.5.3',
}
"""

class QCRetriever:

    def __init__(self):

        self.load_config()
        self.test_peer()

        return

    def load_config(self):

        home_dir = os.environ.get('HOME')
        cfg_json = os.path.join(home_dir, '.__main__.py.json')

        if os.path.isfile(cfg_json):

            # Load config information
            try:
                with os.open(cfg_json, 'r') as fd:
                    cfg = json.load(fd)
            except IOError:
                print('* Could not load config from %s' % cfg_json)
                print('* Exiting')
                sys.exit(1)

    def test_peer(self):

        # Initialise the Application Entity
        ae = AE()

        # Add the storage SCP's supported presentation contexts
        ae.supported_contexts = StoragePresentationContexts

        # Implement the on_c_store callback
        def on_c_store(ds, context, info):
            # Don't store anything, just return Success
            return 0x0000

        # Start the storage SCP on port 11113
        ae.ae_title = b'QC'
        ae.on_c_store = on_c_store
        scp = ae.start_server(('', 11113), block=False)

        # Add a requested presentation context
        ae.add_requested_context(qrc['StudyRootQueryRetrieveInformationModelMove'])

        # Create out identifier (query) dataset
        ds = Dataset()

        ds.QueryRetrieveLevel = 'STUDY'

        # Unique key for PATIENT level
        ds.PatientID = 'QC'

        # Associate with peer AE at IP 127.0.0.1 and port 11112
        # Note: the peer AE must know the IP and port for our move destination
        assoc = ae.associate('131.215.9.86', 11112)

        if assoc.is_established:

            responses = assoc.send_c_move(ds, b'QC', query_model='S')

            # for (status, identifier) in responses:
            #     if status:
            #         print('C-MOVE query status: 0x{0:04x}'.format(status.Status))
            #         if status.Status in (0xFF00, 0xFF01):
            #             print(identifier)
            #     else:
            #         print('Connection timed out, was aborted or received invalid response')

            assoc.release()

        else:

            print('Association rejected, aborted or never connected')

        # Shutdown the storage SCP
        scp.shutdown()



#!/usr/bin/env python3
"""
CBICQC quality control analysis and reporting class
The main analysis and reporting workflow is handled from here

AUTHORS
----
Mike Tyszka, Ph.D., Caltech Brain Imaging Center

DATES
----
2019-05-30 JMT Split out from workflow into single class for easy testing

MIT License

Copyright (c) 2019 Mike Tyszka

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys
import json
import tempfile
import shutil
import numpy as np
import nibabel as nb
import datetime as dt
import pandas as pd
from glob import glob

import bids

from .timeseries import temporal_mean_sd, extract_timeseries, detrend_timeseries
from .graphics import (plot_roi_timeseries, plot_roi_powerspec,
                       plot_mopar_timeseries, plot_mopar_powerspec,
                       orthoslices,
                       roi_demeaned_ts)
from .rois import register_template, make_rois
from .metrics import signal_metrics, moco_metrics
from .moco import moco_phantom, moco_live
from .report import ReportPDF
from .summary import Summarize


class CBICQC:

    def __init__(self, bids_dir, subject='', session='', mode='phantom', past_months=12):

        # Copy arguments into object
        self._bids_dir = bids_dir
        self._subject = subject
        self._session = session
        self._mode = mode
        self._past_months = past_months

        # Phantom or in vivo suffix ('T2star' or 'bold')
        self._suffix = 'T2star' if 'phantom' in mode else 'bold'

        # Batch subject/session ids
        self._this_subject = ''
        self._this_session = ''
        self._this_epits_fpath = ''
        self._this_epits_stub = ''

        # BIDS layout
        self._layout = None

        # Create work and report directories
        self._work_dir = tempfile.mkdtemp()
        self._report_dir = os.path.join(self._bids_dir, 'derivatives', 'cbicqc')
        os.makedirs(self._report_dir, exist_ok=True)

        # Intermediate filenames
        self._report_pdf = ''
        self._report_json = ''
        self._tmean_fname = os.path.join(self._work_dir, 'tmean.nii.gz')
        self._tsd_fname = os.path.join(self._work_dir, 'tsd.nii.gz')
        self._roi_labels_fname = os.path.join(self._work_dir, 'roi_labels.nii.gz')
        self._roi_ts_png = os.path.join(self._work_dir, 'roi_timeseries.png')
        self._roi_ps_png = os.path.join(self._work_dir, 'roi_powerspec.png')
        self._mopar_ts_png = os.path.join(self._work_dir, 'mopar_timeseries.png')
        self._mopar_pspec_png = os.path.join(self._work_dir, 'mopar_powerspec.png')
        self._tmean_montage_png = os.path.join(self._work_dir, 'tmean_montage.png')
        self._tsd_montage_png = os.path.join(self._work_dir, 'tsd_montage.png')
        self._rois_montage_png = os.path.join(self._work_dir, 'rois_montage.png')
        self._rois_demeaned_png = os.path.join(self._work_dir, 'rois_demeaned.png')

        # Flags
        self._save_intermediates = False

        # Metrics of interest to summarize
        self._metrics_df = pd.DataFrame()
        self._metrics_of_interest = []

        # Future proofing. As of pybids version 0.14.0, 'extensio' entity will include the leading dot.
        bids.config.set_option('extension_initial_dot', True)

    def run(self):

        print('')
        print('Starting CBIC EPI Timeseries QC analysis')
        print('')

        # Get complete list of EPI time series for this subject
        # Use _bold.nii.gz suffix to identify time series for now
        # TODO: Broaden search for non-BOLD series
        epits_list = self._get_epits_list()

        # Init list of session metrics for this subject
        metric_list = []

        # Loop over all EPI timeseries images
        for epits_fpath in epits_list:

            self._this_epits_fpath = epits_fpath

            # Parse image filename for BIDS keys
            bids_dict = bids.layout.parse_file_entities(epits_fpath)
            self._this_subject = bids_dict['subject']
            self._this_session = bids_dict['session']

            # Extract EPI timeseries stub from file path
            epits_stub = os.path.basename(epits_fpath).replace('.nii.gz', '')

            print('')
            print('    EPI timeseries {}'.format(epits_stub))

            # Report PDF and JSON filenames - used in both report and summarize modes
            self._report_pdf = os.path.join(
                self._report_dir,
                '{}_qc.pdf'.format(epits_stub)
            )
            self._report_json = self._report_pdf.replace('.pdf', '.json')

            if os.path.isfile(self._report_pdf) and os.path.isfile(self._report_json):

                # QC analysis and reporting already run
                print('      Report and metadata detected for this session')

            else:

                # QC analysis and report generation
                self._analyze_and_report()

            # Add metrics for this subject/session to cumulative list
            metric_list.append(self._get_metrics())

        # Convert metric list to dataframe and save to file
        self._metrics_df = pd.DataFrame(metric_list)

        # Generate summary report for phantom QC only
        if 'phantom' in self._mode:

            # Check for deidentified BIDS data
            # dcm2niix with anonymization on by default generates AcquisitionTime but not AcquisitionDateTime
            # Use "bidskit --no-anon" to skip deidentification and generate AcquisitionDataTime in the JSON sidecars
            if 'AcquisitionDateTime' in self._metrics_df:
                Summarize(self._report_dir, self._metrics_df, self._past_months)
            else:
                print('')
                print('* AcquisitionDateTime is missing from the BIDS metadata')
                print('* The most likely cause of this is deidentification of the data by dcm2niix or bidskit')
                print('* A QC trend summary cannot be generated from deidentified data')


        # Cleanup temporary QC directory
        self.cleanup()

    def _analyze_and_report(self):

        img_fname = self._this_epits_fpath
        json_fname = img_fname.replace('.nii.gz', '.json')

        # Load EPI timeseries image
        print('      Loading EPI timeseries image')
        epits_nii = nb.load(img_fname)

        # Load metadata if available
        print('      Loading QC metadata')
        try:
            with open(json_fname, 'r') as fd:
                meta = json.load(fd)
        except IOError:
            print('      * Could not open image metadata {}'.format(json_fname))
            print('      * Using default imaging parameters')
            meta = self.default_metadata()

        # Check for missing fields (typically non-Siemens scanners)
        if 'SequenceName' not in meta:
            meta['SequenceName'] = 'Unknown Sequence'

        if 'ReceiveCoilName' not in meta:
            meta['ReceiveCoilName'] = 'Unknown Coil'

        if 'BandwidthPerPixelPhaseEncode' not in meta:
            meta['BandwidthPerPixelPhaseEncode'] = '-'

        # Integrate additional meta data from Nifti header and filename
        meta['Subject'] = self._this_subject
        meta['Session'] = self._this_session
        meta['VoxelSize'] = ' x '.join(str(x) for x in epits_nii.header.get('pixdim')[1:4])
        meta['MatrixSize'] = ' x '.join(str(x) for x in epits_nii.shape)

        # Perform rigid body motion correction on EPI timeseries

        # Hardwired moco flag for debugging
        skip_moco = False

        print('      Starting {} motion correction'.format(self._mode))
        t0 = dt.datetime.now()
        epits_moco_nii, epits_moco_pars = self._moco(epits_nii, skip=skip_moco)
        t1 = dt.datetime.now()
        print('      Completed motion correction in {} seconds'.format((t1 - t0).seconds))

        # Temporal mean and sd images
        print('      Calculating temporal mean image')
        tmean_nii, tsd_nii, tsfnr_nii = temporal_mean_sd(epits_moco_nii)

        # Register labels to temporal mean via template image
        print('      Register labels to temporal mean image')
        labels_nii = register_template(tmean_nii, self._work_dir, mode=self._mode)

        # Generate ROIs from labels
        # Construct Nyquist Ghost and airspace ROIs from labels
        rois_nii = make_rois(labels_nii)

        # Extract ROI time series
        print('      Extracting ROI time series')
        s_mean_t = extract_timeseries(epits_moco_nii, rois_nii)

        # Detrend time series
        print('      Detrending time series')
        fit_results, s_detrend_t, s_fit_t = detrend_timeseries(s_mean_t)

        # Calculate signal metrics
        metrics = signal_metrics(fit_results, tsfnr_nii, rois_nii)

        # Add motion metrics to metrics dictionary
        metrics.update(moco_metrics(epits_moco_pars))

        # Add image meta data into metrics dictionary
        metrics.update(meta)

        # Time vector (seconds)
        t = np.arange(0, s_mean_t.shape[1]) * meta['RepetitionTime']

        print('      Generating Report')

        # Create report images
        plot_roi_timeseries(t, s_mean_t, s_detrend_t, s_fit_t, self._roi_ts_png)
        plot_roi_powerspec(t, s_detrend_t, self._roi_ps_png)
        plot_mopar_timeseries(t, epits_moco_pars, self._mopar_ts_png)
        plot_mopar_powerspec(t, epits_moco_pars, self._mopar_pspec_png)
        roi_demeaned_ts(epits_moco_nii, rois_nii, self._rois_demeaned_png)
        orthoslices(tmean_nii, self._tmean_montage_png, cmap='gray', irng='robust')
        orthoslices(tsd_nii, self._tsd_montage_png, cmap='viridis', irng='robust')
        orthoslices(rois_nii, self._rois_montage_png, cmap='Pastel1', irng='noscale')

        # OPTIONAL: Save intermediate images
        if self._save_intermediates:
            nb.save(tmean_nii, self._tmean_fname)
            nb.save(tsd_nii, self._tsd_fname)
            nb.save(rois_nii, self._roi_labels_fname)

        # Construct filename dictionary to pass to PDF generator
        fnames = dict(WorkDir=self._work_dir,
                      ReportPDF=self._report_pdf,
                      ReportJSON=self._report_json,
                      ROITimeseries=self._roi_ts_png,
                      ROIPowerspec=self._roi_ps_png,
                      MoparTimeseries=self._mopar_ts_png,
                      MoparPowerspec=self._mopar_pspec_png,
                      TMeanMontage=self._tmean_montage_png,
                      TSDMontage=self._tsd_montage_png,
                      ROIsMontage=self._rois_montage_png,
                      ROIDemeanedTS=self._rois_demeaned_png,
                      TMean=self._tmean_fname,
                      TSD=self._tsd_fname,
                      ROILabels=self._roi_labels_fname)

        # Build PDF report
        ReportPDF(fnames, meta, metrics)

    def cleanup(self, skip=False):

        if skip:
            print('')
            print('Retaining {}'.format(self._work_dir))
        else:
            print('')
            print('Deleting work directory')
            shutil.rmtree(self._work_dir)

    def _moco(self, img_nii, skip=False):
        """
        Motion correction wrapper

        :param img_nii: Nifti, image object
        :param skip: bool, skip motion correction
        :return moco_nii: Nifti, motion corrected image object
        :return moco_pars: array, motion parameter timeseries
        """

        if skip:

            moco_nii = img_nii
            moco_pars = np.zeros([img_nii.shape[3], 6])

        else:

            if 'phantom' in self._mode:

                moco_nii, moco_pars = moco_phantom(img_nii)

            elif 'live' in self._mode:

                moco_nii, moco_pars = moco_live(img_nii, self._work_dir)

            else:

                print('      * Unknown QC mode ({})'.format(self._mode))
                sys.exit(1)

        return moco_nii, moco_pars

    def _get_metrics(self):

        with open(self._report_json, 'r') as fd:
            metrics = json.load(fd)

        return metrics

    def _get_subject_list(self):

        tmp_list = glob(os.path.join(self._bids_dir, 'sub-*'))
        return [os.path.basename(d).replace('sub-', '') for d in tmp_list]

    def _get_epits_list(self):

        return glob(os.path.join(self._bids_dir, 'sub-*', 'ses-*', 'func', '*_bold.nii.gz'))


    @staticmethod
    def default_metadata():

        meta = dict(Manufacturer='Unkown',
                    Scanner='Unknown',
                    RepetitionTime=3.0,
                    EchoTime=0.030)

        return meta

    @staticmethod
    def parse_filename(fname):

        bname = os.path.basename(fname)

        # Split at underscores
        keyvals = bname.split('_')

        subject, session = 'Unknown', 'Unknown'

        for kv in keyvals:

            if '-' in kv and len(kv) >= 3:

                k, v = kv.split('-')

                if 'sub' in k:
                    subject = v
                if 'ses' in k:
                    session = v

        return subject, session

    @staticmethod
    def save_report(src_pdf, dest_pdf):

        if os.path.isfile(dest_pdf):
            os.remove(dest_pdf)
        elif os.path.isdir(dest_pdf):
            shutil.rmtree(dest_pdf)

        print('Saving QC report to {}'.format(dest_pdf))
        shutil.copyfile(src_pdf, dest_pdf)



# !/usr/bin/env python
"""
CBICQC timeseries analysis node

AUTHOR : Mike Tyszka
PLACE  : Caltech
DATES  : 2019-05-22 JMT From scratch

This file is part of CBICQC.

   CBICQC is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   CBICQC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
  along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2019 California Institute of Technology.
"""

import numpy as np
from scipy.optimize import least_squares
import nibabel as nb


def temporal_mean_sd(qc_moco_nii):

    qc = qc_moco_nii.get_data()

    # Temporal mean of 4D timeseries
    tmean = np.mean(qc, axis=3)
    tsd = np.std(qc, axis=3)
    tsfnr = tmean / (tsd + np.finfo(float).eps)

    tmean_nii = nb.Nifti1Image(tmean, qc_moco_nii.affine)
    tsd_nii = nb.Nifti1Image(tsd, qc_moco_nii.affine)
    tsfnr_nii = nb.Nifti1Image(tsfnr, qc_moco_nii.affine)

    return tmean_nii, tsd_nii, tsfnr_nii


def extract_timeseries(qc_moco_nii, rois_nii):

    rois = rois_nii.get_data()
    s = qc_moco_nii.get_data()

    # Number of time points
    nt = s.shape[3]

    # ROI label indices
    # 0 : unassigned
    # 1 : air space
    # 2 : Nyquist ghost
    # 3 : signal

    labels = [1, 2, 3]
    nl = len(labels)

    s_mean_t = np.zeros([nl, nt])

    for lc in range(0, nl):
        mask = np.array(rois == labels[lc])
        for tc in range(0, nt):
            s_t = s[:, :, :, tc]
            s_mean_t[lc, tc] = np.mean(s_t[mask])

    return s_mean_t


def detrend_timeseries(s_mean_t):
    """
    :param s_mean_t: spatial mean ROI timeseries
    :return:
    """

    # Mean label ROI signal (n_labels x n_timepoints)
    nl, nt = s_mean_t.shape

    # Time vector
    t = np.arange(0, nt)

    s_detrend_t = np.zeros_like(s_mean_t)
    s_fit_t = np.zeros_like(s_mean_t)

    fit_results = []

    # Loop over each label ROI mean timeseries
    for lc in range(0, nl):

        s_t = s_mean_t[lc, :]

        s_min, s_max, s_mean = np.min(s_t), np.max(s_t), np.mean(s_t)
        s_rng = s_max - s_min

        # [Exp Amp, Exp Tau, Linear slope, Offset]
        x0 = [s_rng, 1, -s_rng / float(nt), s_mean]
        bounds = ([0.0, 0, -np.inf, 0],
                  [s_rng, nt, np.inf, np.inf])

        # Robust non-linear curve fit (Huber loss function)
        result = least_squares(explin, x0,
                               method='trf',
                               loss='huber',
                               bounds=bounds,
                               args=(t, s_t))

        # Fitted curve - note sign change from residual definition
        s_fit_t[lc, :] = -explin(result.x, t, 0)

        # Detrended timeseries = y_fit(t) - y(t) + mean(y(t))
        s_detrend_t[lc, :] = result.fun + s_mean

        fit_results.append(result)

    return fit_results, s_detrend_t, s_fit_t


def explin(x, t, y):
    """
    Exponential + linear trend model

    :param x: list, parameters
        0: Exponential amplitude
        1: Exponential time constant
        2: Linear slope
        3: Offset
    :param t: array, time vector
    :param y: array, data
    :return: array, residuals
    """

    y_fit = x[0] * np.exp(-t / x[1]) + x[2] * t + x[3]

    return y - y_fit



# !/usr/bin/env python
"""
Quality control metric calculations from ROI timeseries fit results

AUTHOR : Mike Tyszka
PLACE  : Caltech
DATES  : 2019-06-04 JMT From scratch

This file is part of CBICQC.

   CBICQC is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   CBICQC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
  along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2019 California Institute of Technology.
"""

import numpy as np


def signal_metrics(fit_results, tsfnr_nii, rois_nii):
    """
    Calculate signal QC metrics for each ROI

    :param fit_results: list, fit objects from scipy least_squares
    :return metrics:, dict, QC metric results dictionary
    """

    # TODO:
    # EMI analysis with zipper identification
    # Coil element SNR and fluctuation analysis
    # - Requires separate labels for each coil element

    air_mean = fit_results[0].x[3]

    nyquist_mean = fit_results[1].x[3]

    signal_a_warm = fit_results[2].x[0]
    signal_t_warm = fit_results[2].x[1]
    signal_drift = fit_results[2].x[2]
    signal_mean = fit_results[2].x[3]

    # Calculate main signal tSFNR
    tsfnr = calc_tsfnr(tsfnr_nii, rois_nii)

    # Create and fill dictionary of QC metrics
    metrics = dict()

    metrics['SignalMean'] = signal_mean
    metrics['SNR'] = signal_mean / air_mean
    metrics['SFNR'] = tsfnr
    metrics['SArtR'] = signal_mean / nyquist_mean
    metrics['Drift'] = signal_drift / signal_mean * 100
    metrics['WarmupAmp'] = signal_a_warm / signal_mean * 100
    metrics['WarmupTime'] = signal_t_warm

    # SNR relative to mean noise
    # Estimate spatial noise sigma (assuming underlying Gaussian and Half-Normal distribution)
    # sigma = mean(noise) * sqrt(pi/2)
    # See for example http://en.wikipedia.org/wiki/Half-normal_distribution

    metrics['NoiseSigma'] = air_mean * np.sqrt(np.pi/2)
    metrics['NoiseFloor'] = air_mean
    metrics['SignalSpikes'] = spike_count(fit_results[0].fun)
    metrics['NyquistSpikes'] = spike_count(fit_results[1].fun)
    metrics['AirSpikes'] = spike_count(fit_results[2].fun)

    return metrics


def spike_count(points, thresh=3.5):
    """

    :param points:
    :param thresh:
    :return:
    """

    if len(points.shape) == 1:
        points = points[:, None]

    # Median absolute deviation from the median (MAD)
    med = np.median(points, axis=0)
    dev = np.abs(points - med)
    mad = np.median(dev)

    modified_z_score = 0.6745 * dev / mad

    # Cast to int to prevent JSON encoding errors later
    return int(np.sum(modified_z_score > thresh))


def calc_tsfnr(tsfnr_nii, rois_nii):
    """
    ROI Label Key
    ----
    Undefined       = 0
    Air Space       = 1
    Nyquist Ghost   = 2
    Signal          = 3

    :param tsfnr_nii:
    :param rois_nii:
    :return:
    """

    tsfnr_img = tsfnr_nii.get_data()
    rois_img = rois_nii.get_data()

    return np.mean(tsfnr_img[rois_img == 3])


def moco_metrics(moco_pars):

    # Calculate framewise displacement timecourse from moco_pars
    fd = calc_fd(moco_pars)

    # Create and fill metrics dictionary
    metrics = dict()

    metrics['MeanFD'] = fd.mean()
    metrics['MaxFD'] = fd.max()

    return metrics


def calc_fd(mcf):
    """
    Calculate conventional FD from 6-column FSL MCFLIRT motion parameters

    Reference:
    J. D. Power, K. A. Barnes, A. Z. Snyder, B. L. Schlaggar, and S. E. Petersen,
    “Spurious but systematic correlations in functional connectivity MRI networks arise from subject motion,”
    Neuroimage, vol. 59, pp. 2142–2154, Feb. 2012, doi: 10.1016/j.neuroimage.2011.10.018. [Online].
    Available: http://dx.doi.org/10.1016/j.neuroimage.2011.10.018
    """

    # Rotations (radians)
    rx = mcf[:, 0]
    ry = mcf[:, 1]
    rz = mcf[:, 2]

    # Translations (mm)
    tx = mcf[:, 3]
    ty = mcf[:, 4]
    tz = mcf[:, 5]

    # Backward differences (forward difference + leading 0)

    drx = np.insert(np.diff(rx), 0, 0)
    dry = np.insert(np.diff(ry), 0, 0)
    drz = np.insert(np.diff(rz), 0, 0)

    dtx = np.insert(np.diff(tx), 0, 0)
    dty = np.insert(np.diff(ty), 0, 0)
    dtz = np.insert(np.diff(tz), 0, 0)

    # Total framewise displacement (Power 2012)

    r_sphere = 50.0  # mm

    FD = (np.abs(dtx) +
          np.abs(dty) +
          np.abs(dtz) +
          np.abs(r_sphere * drx) +
          np.abs(r_sphere * dry) +
          np.abs(r_sphere * drz)
          )

    return FD









# !/usr/bin/env python
"""
Graph plotting and image figure functions

AUTHOR : Mike Tyszka
PLACE  : Caltech
DATES  : 2019-05-31 JMT From scratch

This file is part of CBICQC.

   CBICQC is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   CBICQC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
  along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2019 California Institute of Technology.
"""

import numpy as np
import matplotlib.pyplot as plt

from scipy.signal import periodogram
from skimage.util import montage
from skimage.exposure import rescale_intensity
import pandas as pd
from datetime import date

from .moco import total_rotation


def plot_roi_timeseries(t, s_mean_t, s_fit_t, s_detrend_t, plot_fname):
    """
    Plot spatial mean ROI signals vs time

    :param t: float array, time vector (s)
    :param s_mean_t:
    :param s_fit_t:
    :param s_detrend_t:
    :param plot_fname:
    :return:
    """

    roi_names = ['Air', 'Nyquist Ghost', 'Signal']

    fig, axs = plt.subplots(3, 1, figsize=(10, 5))

    for lc in range(0, 3):

        axs[lc].plot(t, s_mean_t[lc, :], label='Raw')
        axs[lc].plot(t, s_detrend_t[lc, :], label='Model')
        axs[lc].plot(t, s_fit_t[lc, :], label='Detrended')
        axs[lc].set_title(roi_names[lc], loc='left')

    # Add x label to final subplot
    axs[2].set_xlabel('Time (s)')

    handles, labels = axs[2].get_legend_handles_labels()
    fig.legend(handles, labels, loc='upper center', ncol=3)

    # Space subplots without title overlap
    plt.tight_layout()

    # Save plot to file
    plt.savefig(plot_fname, dpi=300)

    # Close plot
    plt.close()


def plot_roi_powerspec(t, s_detrend_t, plot_fname):
    """
    Plot ROI timeseries power spectra

    :param t: float array, time vector (s)
    :param s_detrend_t:
    :param plot_fname:
    :return:
    """

    titles = ['Air (dB)', 'Nyquist Ghost (dB)', 'Signal (dB)']

    # Sampling frequency (Hz)
    fs = 1.0 / (t[1] - t[0])

    # Power spectra of detrending residual for each ROI
    f, pspec = periodogram(s_detrend_t, fs, scaling='spectrum')

    # Drop first time point (zeros)
    pspec = pspec[:, 1:]
    f = f[1:]

    plt.subplots(3, 1, figsize=(10, 5))

    for lc in range(0, 3):

        # Power dB relative to row max
        p = pspec[lc, :]
        p_max = np.max(p)
        if p_max < 1e-10:
            p_db = np.zeros_like(p)
        else:
            p_db = 10.0 * np.log10(p / np.max(p))

        plt.subplot(3, 1, lc + 1)
        plt.plot(f, p_db)
        plt.title(titles[lc], loc='left')

    # Add x label to final subplot
    plt.xlabel('Frequency (Hz)')

    # Space subplots without title overlap
    plt.tight_layout()

    # Save plot to file
    plt.savefig(plot_fname, dpi=300)

    # Close plot
    plt.close()


def plot_mopar_timeseries(t, mopars, plot_fname):
    """
    Plots x, y, z displacement and rotation timeseries from MCFLIRT registrations

    mopars columns: (dx, dy, dz, rx, ry, rz)
    Displacements in mm, rotations in degrees

    :param t: float array, nt x 1 time vector
    :param mopars: float array, nt x 6 motion parameters [rx, ry, rz, dx, dy, dz]
    :param plot_fname: str, output plot filename
    :return:
    """

    nt = mopars.shape[0]
    t = np.arange(0, nt)

    plt.subplots(2, 1, figsize=(10, 5))

    plt.subplot(2, 1, 1)
    plt.plot(t, mopars[:, 3:6] * 1e3)
    plt.legend(['x', 'y', 'z'])
    plt.title('Displacement (um)', loc='left')

    plt.subplot(2, 1, 2)
    plt.plot(t, mopars[:, 0:3] * 1e3)
    plt.legend(['x', 'y', 'z'])
    plt.title('Rotation (mdeg)', loc='left')

    # Add x label to final subplot
    plt.xlabel('Time (s)')

    # Space subplots without title overlap
    plt.tight_layout()

    # Save plot to file
    plt.savefig(plot_fname, dpi=300)

    # Close plot
    plt.close()


def plot_mopar_powerspec(t, mopars, plot_fname):
    """
    Plot total motion power spectrum

    :param t: float array, time vector (s)
    :param mopars: array, N x 6
    :param plot_fname:
    :return:
    """

    # Sampling frequency (Hz)
    fs = 1.0 / (t[1] - t[0])

    # Total displacement timecourse
    dtot = np.linalg.norm(mopars[:, 3:6], ord=2, axis=1)

    # Total rotation timecourse
    rtot = total_rotation(mopars[:, 0:3])

    # Total motion array
    mtot = np.array([dtot, rtot])

    # Power spectra of total displacement and rotation timecourses
    f, pspec = periodogram(mtot, fs, scaling='spectrum')

    # Drop first point (zero)
    pspec = pspec[:, 1:]
    f = f[1:]

    plt.subplots(2, 1, figsize=(10, 5))
    titles = ['Displacement (dB)', 'Rotation (dB)']

    for lc in range(0, 2):

        # Power dB relative to row max
        p = pspec[lc, :]
        p_max = np.max(p)
        if p_max < 1e-10:
            p_db = np.zeros_like(p)
        else:
            p_db = 10.0 * np.log10(p / np.max(p))

        plt.subplot(2, 1, lc+1)
        plt.plot(f, p_db)
        plt.title(titles[lc], loc='left')

    # Add x axis label to last subplot
    plt.xlabel('Frequency (Hz)')

    # Space subplots without title overlap
    plt.tight_layout()

    # Save plot to file
    plt.savefig(plot_fname, dpi=300)

    # Close plot
    plt.close()


def orthoslices(img_nii, ortho_fname, cmap='viridis', irng='default'):

    img3d = img_nii.get_data()

    # Intensity scaling
    if 'robust' in irng:
        vmin, vmax = np.percentile(img3d, (1, 99))
    elif 'noscale' in irng:
        nc = plt.get_cmap(cmap).N
        vmin, vmax = 0, nc
    else:
        vmin, vmax = np.min(img3d), np.max(img3d)

    fig, axs = plt.subplots(1, 3, figsize=(7, 2.4), constrained_layout=True)

    nx, ny, nz = img3d.shape
    hx, hy, hz = int(nx/2), int(ny/2), int(nz/2)

    # Extract central section for each orientation
    # Assumes RAS orientation
    m_sag = img3d[hx, :, :].transpose()
    m_cor = img3d[:, hy, :].transpose()
    m_tra = img3d[:, :, hz].transpose()

    # Use transverse image for colorbar reference
    trafig = axs[0].imshow(
        m_tra,
        cmap=plt.get_cmap(cmap),
        vmin=vmin, vmax=vmax,
        aspect='equal',
        origin='lower'
    )
    axs[0].set_title('Transverse')
    axs[0].axis('off')

    axs[1].imshow(
        m_cor,
        cmap=plt.get_cmap(cmap),
        vmin=vmin, vmax=vmax,
        aspect='equal',
        origin='lower'
    )
    axs[1].set_title('Coronal')
    axs[1].axis('off')

    axs[2].imshow(
        m_sag,
        cmap=plt.get_cmap(cmap),
        vmin=vmin, vmax=vmax,
        aspect='equal',
        origin='lower'
    )
    axs[2].set_title('Sagittal')
    axs[2].axis('off')

    plt.colorbar(trafig, ax=axs, location='right', shrink=0.75)

    # Save plot to file
    plt.savefig(ortho_fname, dpi=300)

    # Close plot
    plt.close()

    return ortho_fname


def orthoslice_montage(img_nii, montage_fname, cmap='viridis', irng='default'):

    orient_name = ['Axial', 'Coronal', 'Sagittal']

    img3d = img_nii.get_data()

    plt.subplots(1, 3, figsize=(7, 2.4))

    for ax in [0, 1, 2]:

        # Transpose dimensions for given orientation
        ax_order = np.roll([2, 0, 1], ax)
        s = np.transpose(img3d, ax_order)

        # Downsample to 9 images in first dimension
        nx = s.shape[0]
        xx = np.linspace(0, nx-1, 9).astype(int)
        s = s[xx, :, :]

        # Construct 3x3 montage of slices
        m2d = montage(s, fill='mean', grid_shape=(3, 3))

        # Intensity scaling
        if 'default' in irng:
            m2d = rescale_intensity(m2d, in_range='image', out_range=(0, 1))
        elif 'robust' in irng:
            pmin, pmax = np.percentile(m2d, (1, 99))
            m2d = rescale_intensity(m2d, in_range=(pmin, pmax), out_range=(0, 1))
        else:
            # Do nothing
            pass

        plt.subplot(1, 3, ax+1)
        plt.imshow(m2d,
                   cmap=plt.get_cmap(cmap),
                   aspect='equal',
                   origin='lower')
        plt.title(orient_name[ax])

        plt.axis('off')
        plt.subplots_adjust(bottom=0.0, top=0.9, left=0.0, right=1.0)

    # Remove excess space
    plt.tight_layout()

    # Save plot to file
    plt.savefig(montage_fname, dpi=300)

    # Close plot
    plt.close()


def roi_demeaned_ts(img_nii, rois_nii, residuals_fname):
    """
    Create temporal-spatial image of demeaned voxel timecourse
    - one graymap per ROI
    - subsample voxels in each ROI to yield n_samp timeseries

    :param img_nii:
    :param rois_nii:
    :param residuals_fname: str, ROI residuals PNG filename
    :return:
    """

    # Number of voxel samples from each ROI
    n_samp = 200

    roi_name = ['Air', 'Nyquist Ghost', 'Signal']

    rois = rois_nii.get_data()
    s = img_nii.get_data()

    # Number of time points and labels
    nt = s.shape[3]

    fig, axs = plt.subplots(3, 1, figsize=(7, 9))
    fig.subplots_adjust(left=0.02, bottom=0.06, right=0.95, top=0.94, wspace=0.05)

    for lc in range(3):

        # Skip label 0 - use lc+1 to index ROI labels
        mask3d = np.array(rois == (lc+1))
        mask4d = np.tile(mask3d[:, :, :, np.newaxis], (1, 1, 1, nt))

        nx = np.sum(mask3d)

        s_xt = s[mask4d].reshape(nx, nt)

        # Downsample spatial dimension to n_samp
        inds = np.linspace(0, nx-1, n_samp).astype(int)
        s_xt_d = s_xt[inds, :]

        # Demean rows
        row_means = np.mean(s_xt_d, axis=1)
        res = s_xt_d.astype(np.float64) - np.tile(row_means[:, np.newaxis], (1, nt))

        # Find abs max of residual
        amax = np.abs(res).max()

        # Plot graymap
        pobj = axs[lc].imshow(
            res,
            vmin=-amax,
            vmax=amax,
            cmap=plt.get_cmap('viridis'),
            aspect='auto',
            origin='upper'
        )

        fig.colorbar(pobj, ax=axs[lc])

        axs[lc].set_title(roi_name[lc])
        axs[lc].set_axis_off()

    # Adjust space around plots
    plt.tight_layout()

    # Save plot to file
    plt.savefig(residuals_fname, dpi=300)

    # Close plot
    plt.close()


def metric_trend_plot(mc, metric_name, metrics_df, gridspec, past_months=12):
    """
    Plot session metric trend with median, 5th and 95th percentiles
    Add metric histogram at right

    :param mc: int, index of metric to plot
    :param metric_name: str, metric name to plot
    :param metrics_df: DataFrame, complete metric dataframe for current subject
    :param gridspec: GridSpec, pyplot grid specification
    :param past_months: int, number of past months to plot
    :return:
    """

    # Extract subframe for this metric timeseries
    df = metrics_df[['Date', metric_name, 'Outlier']]

    t1 = pd.Timestamp(date.today())
    t0 = t1 - pd.DateOffset(months=past_months)

    ax0 = plt.subplot(gridspec[mc, 0])

    marker_dict = {'Outlier': 'x', 'Inlier': 'o'}
    color_dict = {'Outlier': 'red', 'Inlier': 'palegreen'}

    for kind in marker_dict:

        inds = df['Outlier'] == kind
        xx = df[inds]['Date']
        yy = df[inds][metric_name]

        ax0.scatter(
            x=xx,
            y=yy,
            c=color_dict[kind],
            ec='black',
        )

    ax0.set_xlim(t0, t1)
    ax0.set_ylabel(metric_name, labelpad=20)
    ax0.xaxis.label.set_size(14)
    ax0.yaxis.label.set_size(14)

    # Calculate metric limits and 5/95 percentiles
    p5, p50, p95 = np.percentile(df[metric_name].values, (5, 50, 95))

    ax0.plot([t0, t1], [p5, p5], 'g:')
    ax0.plot([t0, t1], [p50, p50], 'g')
    ax0.plot([t0, t1], [p95, p95], 'g:')

    ax1 = plt.subplot(gridspec[mc, 1], sharey=ax0)
    df.hist(
        column=metric_name,
        grid=False,
        bins=20,
        orientation='horizontal',
        ec='black',
        fc='palegreen',
        ax=ax1,
    )

    # Simplify graph axes
    ax0.spines['right'].set_visible(False)
    ax0.spines['top'].set_visible(False)
    ax0.yaxis.set_ticks_position('left')
    ax0.xaxis.set_ticks_position('bottom')
    ax1.axis('off')
    ax1.set_title('')
